==> luna_cow-0.5.86-py3-none-any.whl: cow/license/checker.py <==
"""
Implements the process of synchronizing data from the luna-licenses service
"""

import hashlib
from datetime import datetime, timedelta
from hmac import compare_digest
from typing import Any, Awaitable, Optional, Union

from aiosignal import Signal
from apscheduler.job import Job
from apscheduler.schedulers.asyncio import AsyncIOScheduler
from attr import asdict, dataclass
from luna3.common.exceptions import LunaApiException
from luna3.licenses.licenses import LicensesApi
from tzlocal import get_localzone_name
from vlutils.helpers import convertStrFromSnakeToCamelCase, convertTimeToString, convertToSnakeCase

from ..constants.license import LICENSE_TTL
from ..errors.errors import Error, ErrorInfo
from ..errors.exception import VLException
from ..utils import mixins
from ..utils.healthcheck import checkService
from ..utils.log import Logger
from ..utils.rid import generateRequestId
from ..web.error_handlers import createErrorFromLunaApiException


@dataclass(slots=True)
class LicenseFeature:
    """
    Class for storing license feature data
    """

    # (bool): whether the feature available
    isAvailable: bool = False
    # (Optional[Union[int, datetime]]): feature value
    value: Optional[Union[int, str, datetime]] = None

    @classmethod
    def fromDict(cls, data: dict) -> "LicenseFeature":
        """Convert dict to license feature"""
        return cls(isAvailable=data["is_available"], value=data["value"])

    def asDict(self) -> dict:
        """
        Get class dict

        Returns:
            dict with snake_case class members
        """
        featureData = {
            "is_available": self.isAvailable,
            "value": convertTimeToString(self.value, False) if isinstance(self.value, datetime) else self.value,
        }
        return featureData


@dataclass(slots=True)
class LicenseState:
    """
    Class for storing license state
    """

    expirationTime: LicenseFeature | None = None
    facesLimit: LicenseFeature | None = None
    liveness: LicenseFeature | None = None
    livenessBalance: LicenseFeature | None = None
    streamsLimit: LicenseFeature | None = None
    iso: LicenseFeature | None = None
    index: LicenseFeature | None = None
    bodyAttributes: LicenseFeature | None = None
    peopleCounter: LicenseFeature | None = None
    lambdas: LicenseFeature | None = None
    deepfake: LicenseFeature | None = None

    @classmethod
    def fromDict(cls, data: dict) -> "LicenseState":
        """Convert dict to license state"""
        state = cls()
        for key, value in data.items():
            key = convertStrFromSnakeToCamelCase(key)
            setattr(state, key, LicenseFeature.fromDict(value))
        return state

    def asDict(self) -> dict:
        """
        Get class dict

        Returns:
            dict with snake_case class members
        """
        result = {}
        for key, value in convertToSnakeCase(asdict(self, recurse=False)).items():
            if not value:
                continue
            result[key] = value.asDict()
        return result


class LicenseChecker(mixins.Initializable):
    """
    License checker class.

    Attributes:
        licenseClient: luna-license client
        logger: service logger
        scheduler: license check scheduler
        job: checker job
        licenseState: license state
        onLicenseUpdate: signal
        onLicenseError: signal
        targets: list of targets to query from license server, None stands for 'all'
    """

    targets: list[str] | None = None

    def __init__(self, licenseClient: LicensesApi):
        """
        Create license checker instance.
        """
        self.licenseClient = licenseClient
        self.logger = Logger(template="License")
        self.scheduler = AsyncIOScheduler(timezone=get_localzone_name())
        self.job: Job = self.scheduler.add_job(self.reloadLicense, trigger="interval", seconds=LICENSE_TTL)
        self.licenseState: Optional[LicenseState] = None
        self.onLicenseUpdate = Signal(self)
        self.onLicenseError = Signal(self)

        # COMPAT
        self.start = self.initialize
        self.stop = self.close

    def _processLicenseUnavailability(self) -> None:
        """Log and raise error about license unavailability"""
        msg = "Cannot get license information."
        self.logger.error(msg)
        raise VLException(Error.LicenseProblem.format(msg), statusCode=403, isCriticalError=False)

    def checkLicenseAvailability(self) -> bool:
        """Check license availablity"""
        return self.licenseState is not None

    def checkExpirationTime(self) -> bool:
        """Check `expiration_time` availability feature"""
        if not self.checkLicenseAvailability():
            self._processLicenseUnavailability()

        return self.licenseState.expirationTime.isAvailable

    def checkFacesLimit(self) -> bool:
        """Check `faces_limit` availability feature"""
        if not self.checkLicenseAvailability():
            self._processLicenseUnavailability()

        return self.licenseState.facesLimit.isAvailable

    def checkLiveness(self) -> bool:
        """Check `liveness` availability feature"""
        if not self.checkLicenseAvailability():
            self._processLicenseUnavailability()

        return self.licenseState.liveness.isAvailable

    def checklivenessBalance(self) -> bool:
        """Check `liveness_balance` availability feature"""
        if not self.checkLicenseAvailability():
            self._processLicenseUnavailability()

        if self.licenseState.livenessBalance is None:
            return True
        return self.licenseState.livenessBalance.isAvailable

    async def reloadLicense(self) -> None:
        """
        Reload license coro:

        1) try loading new license
        2) if success: end coro
        3) otherwise: pause scheduler and go to "1)"
        """
        try:
            requestId = generateRequestId(self.logger.logTime == "LOCAL")
            reply = await self.licenseClient.getLicense(targets=self.targets, raiseError=True, lunaRequestId=requestId)
            goodSignature = hashlib.pbkdf2_hmac("sha256", reply.body, requestId.encode(), 1000)
            if not compare_digest(goodSignature.hex(), reply.headers.get("Signature", "")):
                error = Error.LicenseProblem.format("Couldn't get license info")
                self.handleError(error)
                await self.onLicenseError.send()
                return
        except LunaApiException as exc:
            error = createErrorFromLunaApiException(exc)
            self.handleError(error)
            await self.onLicenseError.send()
        else:
            self.logger.debug("check succeed")
            self.job.trigger.interval = timedelta(seconds=LICENSE_TTL)
            await self.updateLicenseState(reply.json)
            await self.onLicenseUpdate.send(licenseState=self.licenseState)

    def handleError(self, error: ErrorInfo) -> None:
        """
        Handle load license error

        Args:
            error: vl error
        """
        self.logger.error(error.detail)
        self.logger.error("check failed")
        self.licenseState = None
        self.job.trigger.interval = timedelta(seconds=1)

    async def updateLicenseState(self, licenseResponse: dict[str, Any]) -> None:
        """
        Update license state from luna-licenses data.

        Args:
            licenseResponse: license data
        """
        self.licenseState = LicenseState.fromDict(licenseResponse)

    async def initialize(self):
        """
        Start license checker
        """
        self.onLicenseUpdate.freeze()
        self.onLicenseError.freeze()
        await self.reloadLicense()
        self.scheduler.start()

    async def close(self):
        """
        Stop license checker
        """
        if self.scheduler.running:
            self.scheduler.shutdown()
        await self.licenseClient.close()

    async def probe(self):
        """Check that we can connect to the service with current configuration."""
        return await self.licenseClient.testConnection(totalTimeout=1, asyncRequest=True)

    def getRuntimeChecks(self, includeLunaServices: bool = False) -> list[tuple[str, Awaitable]]:
        """
        Returns configured system checks, pairs of (name, coroutine).

        Args:
            includeLunaServices: A bool, whether to return checks for luna services.
        """
        if not includeLunaServices:
            return []

        return [("licenses", checkService(self.licenseClient, includeLunaServices))]

==> luna_cow-0.5.87-py3-none-any.whl: cow/license/checker.py <==
"""
Implements the process of synchronizing data from the luna-licenses service
"""

import hashlib
from datetime import datetime, timedelta
from hmac import compare_digest
from typing import Any, Awaitable, Optional, Union

from aiosignal import Signal
from apscheduler.job import Job
from apscheduler.schedulers.asyncio import AsyncIOScheduler
from attr import asdict, dataclass
from luna3.common.exceptions import LunaApiException
from luna3.licenses.licenses import LicensesApi
from tzlocal import get_localzone_name
from vlutils.helpers import convertStrFromSnakeToCamelCase, convertTimeToString, convertToSnakeCase

from ..constants.license import LICENSE_TTL
from ..errors.errors import Error, ErrorInfo
from ..errors.exception import VLException
from ..utils import mixins
from ..utils.healthcheck import checkService
from ..utils.log import Logger
from ..utils.rid import generateRequestId
from ..web.error_handlers import createErrorFromLunaApiException


@dataclass(slots=True)
class LicenseFeature:
    """
    Class for storing license feature data
    """

    # (bool): whether the feature available
    isAvailable: bool = False
    # (Optional[Union[int, datetime]]): feature value
    value: Optional[Union[int, str, datetime]] = None

    @classmethod
    def fromDict(cls, data: dict) -> "LicenseFeature":
        """Convert dict to license feature"""
        return cls(isAvailable=data["is_available"], value=data["value"])

    def asDict(self) -> dict:
        """
        Get class dict

        Returns:
            dict with snake_case class members
        """
        featureData = {
            "is_available": self.isAvailable,
            "value": convertTimeToString(self.value, False) if isinstance(self.value, datetime) else self.value,
        }
        return featureData


@dataclass(slots=True)
class LicenseState:
    """
    Class for storing license state
    """

    expirationTime: LicenseFeature | None = None
    facesLimit: LicenseFeature | None = None
    liveness: LicenseFeature | None = None
    livenessBalance: LicenseFeature | None = None
    streamsLimit: LicenseFeature | None = None
    iso: LicenseFeature | None = None
    index: LicenseFeature | None = None
    bodyAttributes: LicenseFeature | None = None
    peopleCounter: LicenseFeature | None = None
    lambdas: LicenseFeature | None = None
    deepfake: LicenseFeature | None = None

    @classmethod
    def fromDict(cls, data: dict) -> "LicenseState":
        """Convert dict to license state"""
        state = cls()
        for key, value in data.items():
            key = convertStrFromSnakeToCamelCase(key)
            setattr(state, key, LicenseFeature.fromDict(value))
        return state

    def asDict(self) -> dict:
        """
        Get class dict

        Returns:
            dict with snake_case class members
        """
        result = {}
        for key, value in convertToSnakeCase(asdict(self, recurse=False)).items():
            if not value:
                continue
            result[key] = value.asDict()
        return result


class LicenseChecker(mixins.Initializable):
    """
    License checker class.

    Attributes:
        licenseClient: luna-license client
        logger: service logger
        scheduler: license check scheduler
        job: checker job
        licenseState: license state
        onLicenseUpdate: signal
        onLicenseError: signal
        targets: list of targets to query from license server, None stands for 'all'
    """

    targets: list[str] | None = None

    def __init__(self, licenseClient: LicensesApi):
        """
        Create license checker instance.
        """
        self.licenseClient = licenseClient
        self.logger = Logger(template="License")
        self.scheduler = AsyncIOScheduler(timezone=get_localzone_name())
        self.job: Job = self.scheduler.add_job(self.reloadLicense, trigger="interval", seconds=LICENSE_TTL)
        self.licenseState: Optional[LicenseState] = None
        self.onLicenseUpdate = Signal(self)
        self.onLicenseError = Signal(self)

        # COMPAT
        self.start = self.initialize
        self.stop = self.close

    def _processLicenseUnavailability(self) -> None:
        """Log and raise error about license unavailability"""
        msg = "Cannot get license information."
        self.logger.error(msg)
        raise VLException(Error.LicenseProblem.format(msg), statusCode=403, isCriticalError=False)

    def checkLicenseAvailability(self) -> bool:
        """Check license availablity"""
        return self.licenseState is not None

    def checkExpirationTime(self) -> bool:
        """Check `expiration_time` availability feature"""
        if not self.checkLicenseAvailability():
            self._processLicenseUnavailability()

        return self.licenseState.expirationTime.isAvailable

    def checkFacesLimit(self) -> bool:
        """Check `faces_limit` availability feature"""
        if not self.checkLicenseAvailability():
            self._processLicenseUnavailability()

        return self.licenseState.facesLimit.isAvailable

    def checkLiveness(self) -> bool:
        """Check `liveness` availability feature"""
        if not self.checkLicenseAvailability():
            self._processLicenseUnavailability()

        return self.licenseState.liveness.isAvailable

    def checklivenessBalance(self) -> bool:
        """Check `liveness_balance` availability feature"""
        if not self.checkLicenseAvailability():
            self._processLicenseUnavailability()

        if self.licenseState.livenessBalance is None:
            return True
        return self.licenseState.livenessBalance.isAvailable

    async def reloadLicense(self) -> None:
        """
        Reload license coro:

        1) try loading new license
        2) if success: end coro
        3) otherwise: pause scheduler and go to "1)"
        """
        try:
            requestId = generateRequestId(self.logger.logTime == "LOCAL")
            reply = await self.licenseClient.getLicense(targets=self.targets, raiseError=True, lunaRequestId=requestId)
            goodSignature = hashlib.pbkdf2_hmac("sha512", reply.body, requestId.encode()[::-1], 100)
            oldSignature = hashlib.pbkdf2_hmac("sha256", reply.body, requestId.encode(), 1000)
            if not compare_digest(goodSignature.hex(), reply.headers.get("Signature", "")):
                if not compare_digest(oldSignature.hex(), reply.headers.get("Signature", "")):
                    error = Error.LicenseProblem.format("Couldn't get license info")
                    self.handleError(error)
                    await self.onLicenseError.send()
                    return
        except LunaApiException as exc:
            error = createErrorFromLunaApiException(exc)
            self.handleError(error)
            await self.onLicenseError.send()
        else:
            self.logger.debug("check succeed")
            self.job.trigger.interval = timedelta(seconds=LICENSE_TTL)
            await self.updateLicenseState(reply.json)
            await self.onLicenseUpdate.send(licenseState=self.licenseState)

    def handleError(self, error: ErrorInfo) -> None:
        """
        Handle load license error

        Args:
            error: vl error
        """
        self.logger.error(error.detail)
        self.logger.error("check failed")
        self.licenseState = None
        self.job.trigger.interval = timedelta(seconds=1)

    async def updateLicenseState(self, licenseResponse: dict[str, Any]) -> None:
        """
        Update license state from luna-licenses data.

        Args:
            licenseResponse: license data
        """
        self.licenseState = LicenseState.fromDict(licenseResponse)

    async def initialize(self):
        """
        Start license checker
        """
        self.onLicenseUpdate.freeze()
        self.onLicenseError.freeze()
        await self.reloadLicense()
        self.scheduler.start()

    async def close(self):
        """
        Stop license checker
        """
        if self.scheduler.running:
            self.scheduler.shutdown()
        await self.licenseClient.close()

    async def probe(self):
        """Check that we can connect to the service with current configuration."""
        return await self.licenseClient.testConnection(totalTimeout=1, asyncRequest=True)

    def getRuntimeChecks(self, includeLunaServices: bool = False) -> list[tuple[str, Awaitable]]:
        """
        Returns configured system checks, pairs of (name, coroutine).

        Args:
            includeLunaServices: A bool, whether to return checks for luna services.
        """
        if not includeLunaServices:
            return []

        return [("licenses", checkService(self.licenseClient, includeLunaServices))]
